from pathlib import Path
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn

from gwas_sumstats_tools.interfaces.metadata import (
    MetadataClient,
    metadata_dict_from_gwas_cat,
    get_file_metadata,
)
from gwas_sumstats_tools.utils import (
    parse_accession_id,
    append_to_path,
    exit_if_no_data,
)

class Gen_meta:
    def __init__(
        self,
        data_infile: Path,
        metadata_infile: Path = None,
        metadata_outfile: Path = None,
        format_data: bool = False,
    ) -> None:
        self.format_data = format_data
        self.data_infile = Path(data_infile)
        self.metadata_outfile = Path(
            self._set_metadata_outfile_name()
            if not metadata_outfile
            else metadata_outfile
        )
        self.meta = MetadataClient(
            in_file=metadata_infile, out_file=self.metadata_outfile
        )

    def _set_metadata_outfile_name(self) -> str:
        #return append_to_path(self.data_outfile, "-meta.yaml")
        return append_to_path(self.data_infile, "-meta.yaml")

    def set_metadata(
        self, from_gwas_cat: bool = False, custom_metadata: dict = None
    ) -> MetadataClient:
        """Set metadata.
        The hierarchy of where metadata is set is as follows:
        1. custom_metadata map (overwrites anything below)
        2. metadata from the GWAS Catalog
        3. metadata from the original input file
        4. metadata inferred from the datafile

        Keyword Arguments:
            from_gwas_cat -- update with data from GWAS catalog (default: {False})
            custom_metadata -- Update with this map (default: {None})

        Returns:
            metadata object
        """

        meta_dict = get_file_metadata(
            in_file=self.data_infile,
            out_file=self.data_infile,
            meta_dict=self.meta.as_dict(),
        )
        if from_gwas_cat:
            accession_id = parse_accession_id(filename=self.data_infile)
            meta_dict.update(metadata_dict_from_gwas_cat(accession_id=accession_id))
        if custom_metadata:
            meta_dict.update(custom_metadata)
        self.meta.update_metadata(meta_dict)
        return self.meta


def gen_meta(
    filename: Path,
    generate_metadata: bool = False,
    metadata_outfile: Path = None,
    metadata_infile: Path = None,
    metadata_from_gwas_cat: bool = False,
    metadata_dict: dict = None,
) -> None:
    gen_meta = Gen_meta(
        data_infile=filename,
        metadata_infile=metadata_infile,
        metadata_outfile=metadata_outfile,
    )
    # Get metadata
    
    print("[bold]\n---------- METADATA ----------\n[/bold]")
    metadata = gen_meta.set_metadata(
        from_gwas_cat=metadata_from_gwas_cat, custom_metadata=metadata_dict
        )
    print(metadata)
    print(f"[green]Writing metadata --> {str(metadata_outfile)}[/green]")
    metadata.to_file()


import yaml
import json
from typing import Union
from datetime import date
from pathlib import Path
import ruamel.yaml

from pydantic import ValidationError
from gwas_sumstats_tools.config import (GWAS_CAT_API_STUDIES_URL,
                                        GWAS_CAT_API_INGEST_STUDIES_URL,
                                        GWAS_CAT_API_INGEST_STUDIES_URL_SANDBOX,
                                        GWAS_CAT_STUDY_MAPPINGS,
                                        REST_GWAS_CAT_STUDY_MAPPINGS,
                                        STUDY_FIELD_TO_SPLIT,
                                        SAMPLE_FIELD_TO_SPLIT,
                                        GWAS_CAT_SAMPLE_MAPPINGS,
                                        REST_GWAS_CAT_SAMPLE_MAPPINGS,
                                        GENOME_ASSEMBLY_MAPPINGS)
from gwas_sumstats_tools.utils import (download_with_requests,
                                       parse_accession_id,
                                       parse_genome_assembly,
                                       get_md5sum,
                                       replace_dictionary_keys,
                                       split_fields_on_delimiter,
                                       update_dict_if_not_set)
from gwas_sumstats_tools.schema.metadata import SumStatsMetadata


class MetadataClient:
    def __init__(self, meta_dict: dict = {},
                 in_file: Path = None,
                 out_file: Path = None) -> None:
        """SumStats Metadata client

        Keyword Arguments:
            meta_dict -- Dict of metadata (default: {None})
            in_file -- Input metadata YAML file (default: {None})
            out_file -- Output metadata YAML file (default: {None})
        """
        self.metadata = SumStatsMetadata.construct()
        self._meta_dict = meta_dict
        self._in_file = in_file
        self._out_file = out_file

    def from_file(self) -> None:
        """Create metadata from YAML file
        """
        if self._in_file is None:
            return None
        with open(self._in_file, "r") as fh:
            self._meta_dict = yaml.safe_load(fh)
            self.update_metadata(self._meta_dict)

    def to_file(self) -> None:
        """Write metadata to YAML file
        """
        yaml_data =yaml.dump(self.metadata.dict(exclude_none=True), default_flow_style=False,sort_keys=False)
        yaml_yaml=ruamel.yaml.YAML().load(str(yaml_data))
        yaml_yaml.yaml_set_start_comment("Study meta-data")
        yaml_yaml.yaml_set_comment_before_after_key('trait_description','\nTrait Information')
        yaml_yaml.yaml_set_comment_before_after_key('genome_assembly','\nGenotyping Information')
        yaml_yaml.yaml_set_comment_before_after_key('samples','\nSample Information')
        yaml_yaml.yaml_set_comment_before_after_key('data_file_name','\nSummary Statistic information')
        yaml_yaml.yaml_set_comment_before_after_key('is_harmonised','\nHarmonization status')

        with open(self._out_file, "w") as fh:
            yml = ruamel.yaml.YAML()
            yml.indent(mapping=2, sequence=4, offset=2)
            yml.dump(yaml_yaml,
                      fh)

    def update_metadata(self, data_dict: dict) -> None:
        """Create a copy of the model and update (no validation occurs)

        Arguments:
            data_dict -- Dict of data to populate model
        """
        self._meta_dict.update(data_dict)
        try:
            self.metadata = self.metadata.parse_obj(self._meta_dict)
        except ValidationError as e:
            print(f"Metadata not updated due to the following error:\n {e}\n")

    def __repr__(self) -> str:
        """Representation of metadata.
        """
        return self.as_yaml()

    def as_dict(self) -> dict:
        """Dict repr of metadata

        Returns:
            Dict of metadata
        """
        return self.metadata.dict()

    def as_yaml(self, **kwargs) -> str:
        return yaml.dump(self.metadata.dict(**kwargs),sort_keys=False,default_flow_style=False)
    
    def yscbak(self, key, before=None, indent=0, after=None, after_indent=None):
        """
        expects comment (before/after) to be without `#` and possible have multiple lines
        """
        from ruamel.yaml.error import Mark
        from ruamel.yaml.tokens import CommentToken
        
        def comment_token(s, mark):
            # handle empty lines as having no comment
            return CommentToken(('# ' if s else '') + s + '\n', mark, None)
        
        if after_indent is None:
            after_indent = indent + 2
        if before and before[-1] == '\n':
            before = before[:-1]  # strip final newline if there
        if after and after[-1] == '\n':
            after = after[:-1]  # strip final newline if there
        start_mark = Mark(None, None, None, indent, None, None)
        c = self.ca.items.setdefault(key, [None, [], None, None])
        if before:
            for com in before.split('\n'):
                c[1].append(comment_token(com, start_mark))
        if after:
            start_mark = Mark(None, None, None, after_indent, None, None)
            if c[3] is None:
                c[3] = []
            for com in after.split('\n'):
                c[3].append(comment_token(com, start_mark))
    if not hasattr(ruamel.yaml.comments.CommentedMap,'yaml_set_comment_before_after_key'):
        ruamel.yaml.comments.CommentedMap.yaml_set_comment_before_after_key = yscbak

def metadata_dict_from_gwas_cat(
    accession_id: str, 
    is_bypass_rest_api: bool = False, 
    is_sandbox: bool = False,
) -> dict:
    """Extract metadat from the GWAS Catalog API

    Arguments:
        accession_id(str): GWAS Catalog accession ID
        is_bypass_rest_api(bool, optional): A flag indicating whether the request comes from the sumstats
            service. If True, the function bypasses querying the REST API for additional study metadata.
            Defaults to False, which means the REST API response will be included in the metadata dict.
        is_sandbox(bool, optional): A A flag indicating whether the request comes from the sandbox sumstats
            service.

    Returns:
        Metadata dict
    """
    meta_dict = {}
    sample_list = []
    study_url = GWAS_CAT_API_INGEST_STUDIES_URL + accession_id
    if is_sandbox:
        study_url = GWAS_CAT_API_INGEST_STUDIES_URL_SANDBOX + accession_id
    sample_url = study_url + "/samples"
    rest_url = GWAS_CAT_API_STUDIES_URL + accession_id

    study_response = download_with_requests(url=study_url)
    sample_response = download_with_requests(url=sample_url, params={"size": 100})

    if not is_bypass_rest_api:
        rest_response = download_with_requests(url=rest_url)

        try:
            if rest_response:
                print(f"{rest_url} returned 200")
            rest_dict = _parse_gwas_rest_study_response(
                rest_response,
                replace_dict=GWAS_CAT_STUDY_MAPPINGS,
                fields_to_split=STUDY_FIELD_TO_SPLIT,
            )
            meta_dict.update(rest_dict)
        except Exception as e:
            print(f"Error processing REST API response: {e}")
            pass
    else:
        try:
            rest_dict = _parse_ingest_study_response(
                study_response,
                replace_dict=GWAS_CAT_STUDY_MAPPINGS,
                fields_to_split=STUDY_FIELD_TO_SPLIT,
            )
            print(f"Ingest Study Response ::: {rest_dict=}")
            meta_dict.update(rest_dict)
        except Exception as e:
            print(f"Error processing REST API response: {e}")
            pass

    try:
        ingest_dict = _parse_gwas_api_study_response(
            study_response,
            replace_dict=GWAS_CAT_STUDY_MAPPINGS,
            fields_to_split=STUDY_FIELD_TO_SPLIT,
        )
        meta_dict.update(ingest_dict)

        # Update trait_description
        d = meta_dict.get("trait_description")
        t = meta_dict.get("trait")
        if not d and t:
            meta_dict.update({"trait_description": t})
    except Exception as e:
        print(f"Error processing Ingest API response: {e}")
        pass

    try:
        ingest_samples_list = _parse_gwas_api_samples_response(
            sample_response,
            replace_dict=GWAS_CAT_SAMPLE_MAPPINGS,
            fields_to_split=SAMPLE_FIELD_TO_SPLIT,
        )
        sample_list = ingest_samples_list
    except Exception as e:
        print(f"Error processing Ingest Samples API response: {e}")
        pass

    if not sample_list:
        try:
            rest_samples_list = _parse_gwas_rest_samples_response(
                rest_response,
                replace_dict=GWAS_CAT_SAMPLE_MAPPINGS,
                fields_to_split=SAMPLE_FIELD_TO_SPLIT,
            )
            sample_list = rest_samples_list
        except Exception as e:
            print(
                f"Ample info of {accession_id} is missing in the REST API and INGEST API: {e}"
            )

    meta_dict["samples"] = sample_list
    return meta_dict


def _parse_gwas_api_study_response(response: bytes,
                                   replace_dict: dict = None,
                                   fields_to_split: tuple = None
                                   ) -> dict:
    """Parse study repsonse from GWAS cat api

    Arguments:
        response -- response bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        Dict of metadata
    """
    result_dict = {}
    if response:
        result_dict = json.loads(response.decode())
        if replace_dict:
            result_dict = replace_dictionary_keys(data_dict=result_dict,
                                                  replace_dict=replace_dict)
        if fields_to_split:
            result_dict = split_fields_on_delimiter(data_dict=result_dict,
                                                   fields=fields_to_split)
    return result_dict


def _parse_ingest_study_response(
    response: bytes,
    replace_dict: dict = None,
    fields_to_split: tuple = None
) -> dict:
    response_parsed = {}
    response = json.loads(response.decode())

    # study
    trait_description = response.get("diseaseTrait", {}).get("trait")
    if trait_description:
        response_parsed["trait_description"] = [trait_description]

    # # extract EFO
    # efo_url=response['_links']['efoTraits']['href']
    # efo_response =download_with_requests(url=efo_url)
    # if efo_response:
    #     efo_info=json.loads(efo_response.decode())['_embedded']['efoTraits']
    #     response_parsed["ontology_mapping"]="|".join(d.get("shortForm") for d in efo_info)
    # if replace_dict:
    #     response_parsed = replace_dictionary_keys(data_dict=response_parsed,
    #                                               replace_dict=replace_dict)
    # if fields_to_split:
    #     response_parsed = split_fields_on_delimiter(data_dict=response_parsed,
    #                                                 fields=fields_to_split)
    return response_parsed


def _parse_gwas_rest_study_response(response: bytes,
                                   replace_dict: dict = None,
                                   fields_to_split: tuple = None
                                   ) -> dict:
    """Parse study repsonse from GWAS cat rest api

    Arguments:
        response -- response bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        Dict of metadata
    """
    result_dict={}
    response_dict = json.loads(response.decode())
    # study
    result_dict["trait_description"]=response_dict['diseaseTrait'].get('trait')
    result_dict["genotyping_technology"]="|".join(d.get("genotypingTechnology") for d in response_dict['genotypingTechnologies'])
    
    # extract EFO
    efo_url=response_dict['_links']['efoTraits']['href']
    efo_response =download_with_requests(url=efo_url)
    if efo_response:
        efo_info=json.loads(efo_response.decode())['_embedded']['efoTraits']
        result_dict["ontology_mapping"]="|".join(d.get("shortForm") for d in efo_info)
    if replace_dict:
        result_dict = replace_dictionary_keys(data_dict=result_dict,
                                                  replace_dict=replace_dict)
    if fields_to_split:
        result_dict = split_fields_on_delimiter(data_dict=result_dict,
                                                    fields=fields_to_split)
    return result_dict


def _parse_gwas_api_samples_response(response: bytes,
                                     replace_dict: dict = None,
                                     fields_to_split: tuple = None
                                     ) -> list:
    """Parse the samples response from GWAS cat api

    Arguments:
        response -- responce bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        List of samples dicts
    """
    formatted_list = []
    if response:
        result_dict = json.loads(response.decode())
        sample = result_dict["_embedded"].get('samples')
        sample_list=[x for x in sample if x.get('stage') == 'discovery']
        if sample_list:
            for element in sample_list:
                if replace_dict:
                    element = replace_dictionary_keys(data_dict=element,
                                                      replace_dict=replace_dict)
                if fields_to_split:
                    element = split_fields_on_delimiter(data_dict=element,
                                                        fields=fields_to_split)
                formatted_list.append(element)
    return formatted_list

def _parse_gwas_rest_samples_response(response: bytes,
                                     replace_dict: dict = None,
                                     fields_to_split: tuple = None
                                     ) -> list:
    """Parse the samples response from GWAS cat api

    Arguments:
        response -- responce bytes

    Keyword Arguments:
        replace_dict -- Header mappings (default: {None})
        fields_to_split -- fields to split (default: {None})

    Returns:
        List of samples dicts
    """
    response_dict = json.loads(response.decode())
    #case_control
    sample_dict={}
    if ("case" in response_dict['initialSampleSize'] ) and ("control" in response_dict['initialSampleSize']):
        sample_dict["case_control_study"]="true"
    formatted_list = []
    ancestry=response_dict['ancestries']
    sample_list=[x for x in ancestry if x.get('type') == 'initial']
    if sample_list:
        for element in sample_list:
            element=dict((k,element[k]) for k in REST_GWAS_CAT_SAMPLE_MAPPINGS.keys() if k in element)
            element['ancestralGroups']=[x.get('ancestralGroup') for x in element['ancestralGroups']]
            element.update(sample_dict)
            
            if replace_dict:
                element=replace_dictionary_keys(element,REST_GWAS_CAT_SAMPLE_MAPPINGS)
            if fields_to_split:
                element=split_fields_on_delimiter(element,REST_GWAS_CAT_SAMPLE_MAPPINGS)
            formatted_list.append(element)

    return formatted_list

def get_file_metadata(in_file: Path, out_file: str, meta_dict: dict = {}) -> dict:
    """Get file related metadata

    Arguments:
        in_file -- sumstats in file
        outfile -- sumstats out file

    Returns:
        Metadata dict
    """
    inferred_meta_dict = {}
    inferred_meta_dict['gwas_id'] = parse_accession_id(filename=in_file)
    inferred_meta_dict['data_file_name'] = Path(out_file).name
    inferred_meta_dict['file_type'] = 'GWAS-SSF v1.0'
    inferred_meta_dict['genome_assembly'] = GENOME_ASSEMBLY_MAPPINGS.get(parse_genome_assembly(filename=in_file), 'unknown')
    inferred_meta_dict['data_file_md5sum'] = get_md5sum(out_file) if Path(out_file).exists() else None
    inferred_meta_dict['date_metadata_last_modified'] = date.today()
    inferred_meta_dict['gwas_catalog_api'] = GWAS_CAT_API_STUDIES_URL + parse_accession_id(filename=in_file)
    for field, value in inferred_meta_dict.items():
        update_dict_if_not_set(meta_dict, field, value)
    return meta_dict


def init_metadata_from_file(filename: Path, metadata_infile: Path = None) -> Union[SumStatsMetadata, None]:
    m_in = metadata_infile if metadata_infile else filename.with_suffix(filename.suffix + "-meta.yaml")
    if m_in.exists():
        ssm = MetadataClient(in_file=m_in)
        ssm.from_file()
        return ssm
    else:
        return None



from pathlib import Path
from typing import List, Optional
import typer
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn

from gwas_sumstats_tools.gen_meta import gen_meta
from gwas_sumstats_tools.validate import validate
from gwas_sumstats_tools.read import read
from gwas_sumstats_tools.format import format
from gwas_sumstats_tools.utils import (header_dict_from_args,
                                       metadata_dict_from_args,
                                       get_version)


app = typer.Typer(add_completion=False,
                  no_args_is_help=True,
                  rich_markup_mode="rich",
                  context_settings={"help_option_names": ["-h", "--help"]})


def exit_status(status: bool) -> int:
    return 0 if status is True else 1


@app.command("validate",
             no_args_is_help=True,
             context_settings={"help_option_names": ["-h", "--help"],
                               "allow_extra_args": True,
                               "ignore_unknown_options": True})
def ss_validate(filename: Path = typer.Argument(...,
                                                exists=True,
                                                readable=True,
                                                help="Input sumstats file. Must be TSV or CSV and may be gzipped"),
                errors_file: bool = typer.Option(False,
                                                 "--errors-out", "-e",
                                                 help="Output erros to a csv file, <filename>.err.csv.gz"),
                pval_zero: bool = typer.Option(False,
                                               "--p-zero", "-z",
                                               help="Force p-values of zero to be allowable. Takes precedence over inferred value (-i)"),
                minimum_rows: int = typer.Option(100_000,
                                                 "--min-rows", "-m",
                                                 help="Minimum rows acceptable for the file"),
                chunkzize: int = typer.Option(1_000_000,
                                              "--chunksize", "-s",
                                              help=("Number of rows to store in memory at once. "
                                                    "Increase this number for more speed at the cost "
                                                    "of more memory. Decrease to save memory, at the "
                                                    "cost of speed.")),
                infer_from_metadata: bool = typer.Option(False,
                                                         "--infer-from-metadata", "-i",
                                                         help=("Infer validation options from the "
                                                               "metadata file <filename>-meta.yaml. "
                                                               "E.g. fields for analysis software and "
                                                               "negative log10 p-values affect the data "
                                                               "validation behaviour."))
                ):
    """
    [green]VALIDATE[/green] a GWAS summary statistics data file
    """
    print(f"Validating file: {filename}")
    with Progress(SpinnerColumn(),
                  TextColumn("[progress.description]{task.description}"),
                  transient=True
                  ) as progress:
        progress.add_task(description="Validating...", total=None)
        (valid,
         message,
         error_preview,
         error_type) = validate(filename=filename,
                                errors_file=errors_file,
                                pval_zero=pval_zero,
                                minimum_rows=minimum_rows,
                                chunksize=chunkzize,
                                infer_from_metadata=infer_from_metadata)
    print(f"Validation status: {valid}")
    print(message)
    if error_type:
        print(("Primary reason for validation failure: "
               f"[red]{error_type}[/red]"))
    if error_preview:
        print(("See below for a preview of the errors. "
               "To get all the errors in a file run the "
               "[green][bold]validate[/bold][/green] command "
               "with the [green][bold]-e[/bold][/green] flag."))
        print(error_preview)
    raise typer.Exit(exit_status(valid))


@app.command("read",
             no_args_is_help=True,
             context_settings={"help_option_names": ["-h", "--help"],
                               "allow_extra_args": True,
                               "ignore_unknown_options": True})
def ss_read(filename: Path = typer.Argument(...,
                                            exists=True,
                                            readable=True,
                                            help="Input sumstats file. Must be TSV or CSV and may be gzipped"),
            get_header: bool = typer.Option(False,
                                            "--get-header", "-h",
                                            help="Just return the headers of the file"),
            metadata_infile: Path = typer.Option(None,
                                                 "--meta-in",
                                                 readable=True,
                                                 exists=True,
                                                 help=("Specify a metadata file to read in, "
                                                       "defaulting to <filename>-meta.yaml"),
                                                 show_default=False),
            get_all_metadata: bool = typer.Option(False,
                                                  "--get-all-metadata", "-M",
                                                  help="Return all metadata"),
            get_metadata: Optional[List[str]] = typer.Option(None,
                                                             "--get-metadata", "-m",
                                                             help=("Get metadata for the "
                                                                   "specified fields e.g. "
                                                                   "`-m genomeAssembly -m isHarmonised"))
            ):
    """
    [green]READ[/green] a sumstats file
    """
    result, message = read(filename=filename,
                           metadata_infile=metadata_infile,
                           get_header=get_header,
                           get_all_metadata=get_all_metadata,
                           get_metadata=get_metadata)
    print(message)
    print(result)


@app.command("format",
             no_args_is_help=True,
             context_settings={"help_option_names": ["-h", "--help"],
                               "allow_extra_args": True,
                               "ignore_unknown_options": True})
def ss_format(filename: Path = typer.Argument(...,
                                              readable=True,
                                              help="Input sumstats file. Must be TSV or CSV and may be gzipped"),
              data_outfile: Path = typer.Option(None,
                                                "--ss-out", "-o",
                                                writable=True,
                                                file_okay=True,
                                                help="Output sumstats file"),
              minimal_to_standard: bool = typer.Option(False,
                                                 "--minimal2standard", "-s",
                                                 help=("Try to convert a valid, minimally formatted file "
                                                       "to the standard format."
                                                       "This assumes the file at least has `p_value` "
                                                       " combined with rsid in `variant_id` field or "
                                                       "`chromosome` and `base_pair_location`. Validity "
                                                       "of the new file is not guaranteed because mandatory "
                                                       "data could be missing from the original file.")),
              generate_metadata: bool = typer.Option(False,
                                                     "--generate-metadata", "-m",
                                                     help="Create the metadata file"),
              metadata_outfile: Path = typer.Option(None,
                                                    "--meta-out",
                                                    writable=True,
                                                    file_okay=True,
                                                    help="Specify the metadata output file"),
              metadata_infile: Path = typer.Option(None,
                                                   "--meta-in",
                                                   readable=True,
                                                   exists=True,
                                                   help="Specify a metadata file to read in"),
              metadata_edit_mode: bool = typer.Option(False,
                                                      "--meta-edit", "-e",
                                                      help=("Enable metadata edit mode. "
                                                            "Then provide params to edit in "
                                                            "the `--<FIELD>=<VALUE>` format e.g. "
                                                            "`--GWASID=GCST123456` to edit/add "
                                                            "that value")),
              metadata_from_gwas_cat: bool = typer.Option(False,
                                                          "--meta-gwas", "-g",
                                                          help="[italic]Internal use only[/italic]. Populate metadata from GWAS Catalog"),
              custom_header_map: bool = typer.Option(False,
                                                     "--custom-header-map", "-c",
                                                     help=("Provide a custom header mapping using "
                                                           "the `--<FROM>:<TO>` format e.g. "
                                                           "`--chr:chromosome`")),
              extra_args: typer.Context = typer.Option(None)
              ):
    """
    [green]FORMAT[/green] a sumstats file by creating a new one from the existing one. Add/edit metadata.
    """
    header_map = header_dict_from_args(args=extra_args.args) \
        if custom_header_map else {}
    meta_dict = metadata_dict_from_args(args=extra_args.args) \
        if metadata_edit_mode else {}
    format(filename=filename,
           data_outfile=data_outfile,
           minimal_to_standard=minimal_to_standard,
           generate_metadata=generate_metadata,
           metadata_outfile=metadata_outfile,
           metadata_infile=metadata_infile,
           metadata_from_gwas_cat=metadata_from_gwas_cat,
           header_map=header_map,
           metadata_dict=meta_dict)

@app.command("gen_meta",
             no_args_is_help=True,
             context_settings={"help_option_names": ["-h", "--help"],
                               "allow_extra_args": True,
                               "ignore_unknown_options": True})
def ss_gen_meta(filename: Path = typer.Argument(...,
                                              readable=True,
                                              help="Input sumstats file. Must be TSV or CSV and may be gzipped"),
              generate_metadata: bool = typer.Option(False,
                                                     "--generate-metadata", "-m",
                                                     help="Create the metadata file"),
              metadata_outfile: Path = typer.Option(None,
                                                    "--meta-out",
                                                    writable=True,
                                                    file_okay=True,
                                                    help="Specify the metadata output file"),
              metadata_infile: Path = typer.Option(None,
                                                   "--meta-in",
                                                   readable=True,
                                                   exists=True,
                                                   help="Specify a metadata file to read in"),
              metadata_edit_mode: bool = typer.Option(False,
                                                      "--meta-edit", "-e",
                                                      help=("Enable metadata edit mode. "
                                                            "Then provide params to edit in "
                                                            "the `--<FIELD>=<VALUE>` format e.g. "
                                                            "`--GWASID=GCST123456` to edit/add "
                                                            "that value")),
              metadata_from_gwas_cat: bool = typer.Option(False,
                                                          "--meta-gwas", "-g",
                                                          help="[italic]Internal use only[/italic]. Populate metadata from GWAS Catalog"),
              extra_args: typer.Context = typer.Option(None)
              ):
    """
    [green]FORMAT[/green] a sumstats file by creating a new one from the existing one. Add/edit metadata.
    """
    meta_dict = metadata_dict_from_args(args=extra_args.args) \
        if metadata_edit_mode else {}
    gen_meta(filename=filename,
           generate_metadata=generate_metadata,
           metadata_outfile=metadata_outfile,
           metadata_infile=metadata_infile,
           metadata_from_gwas_cat=metadata_from_gwas_cat,
           metadata_dict=meta_dict)


@app.command("version")
def ss_version():
    """Print installed version and exit.
    """
    print(get_version())


if __name__ == "__main__":
    app()
